import asyncio

from blaxel.agents import bl_agent


async def main():
    agent = bl_agent("template-google-adk-py")
    print(await agent.arun({"inputs": "Hello, world!"}))
    print(agent.run({"inputs": "Hello, world!"}))

if __name__ == "__main__":
    asyncio.run(main())



import asyncio
from typing import Dict

from ..common.settings import settings
from .base import SandboxHandleBase
from .client.api.filesystem.delete_filesystem_path import (
    asyncio_detailed as delete_filesystem_by_path,
)
from .client.api.filesystem.get_filesystem_path import asyncio_detailed as get_filesystem_by_path
from .client.api.filesystem.put_filesystem_path import asyncio_detailed as put_filesystem_by_path
from .client.client import client
from .client.models import Directory, FileRequest, SuccessResponse


class SandboxFileSystem(SandboxHandleBase):
    def __init__(self, sandbox):
        super().__init__(sandbox)
        self.client = client.with_base_url(self.url).with_headers(settings.headers)

    async def mkdir(self, path: str, permissions: str = "0755") -> SuccessResponse:
        path = self.format_path(path)
        body = FileRequest(is_directory=True, permissions=permissions)
        response = await put_filesystem_by_path(path=path, client=self.client, body=body)
        self.handle_response(response)
        return response.parsed

    async def write(self, path: str, content: str) -> SuccessResponse:
        path = self.format_path(path)
        body = FileRequest(content=content)
        response = await put_filesystem_by_path(path=path, client=self.client, body=body)
        self.handle_response(response)
        return response.parsed

    async def read(self, path: str) -> str:
        path = self.format_path(path)
        response = await get_filesystem_by_path(path=path, client=self.client)
        self.handle_response(response)
        if "content" not in response.parsed.additional_properties:
            raise Exception('{"error": "File not found"}')
        return response.parsed.additional_properties["content"]

    async def rm(self, path: str, recursive: bool = False) -> SuccessResponse:
        path = self.format_path(path)
        response = await delete_filesystem_by_path(path=path, client=self.client, recursive=recursive)
        self.handle_response(response)
        return response.parsed

    async def ls(self, path: str) -> Directory:
        path = self.format_path(path)
        response = await get_filesystem_by_path(path=path, client=self.client)
        self.handle_response(response)
        if not hasattr(response.parsed, "files") and not hasattr(response.parsed, "subdirectories"):
            raise Exception('{"error": "Directory not found"}')
        return response.parsed

    async def cp(self, source: str, destination: str) -> Dict[str, str]:
        source = self.format_path(source)
        destination = self.format_path(destination)
        response = await get_filesystem_by_path(path=source, client=self.client)
        self.handle_response(response)
        data = response.parsed
        if "content" in data.additional_properties:
            await self.write(destination, data.additional_properties["content"])
            return {
                "message": "File copied successfully",
                "source": source,
                "destination": destination,
            }
        elif hasattr(data, "subdirectories") or hasattr(data, "files"):
            # Create destination directory
            await self.mkdir(destination)
            # Process subdirectories in batches of 5
            subdirectories = getattr(data, "subdirectories", []) or []
            for i in range(0, len(subdirectories), 5):
                batch = subdirectories[i:i+5]
                await asyncio.gather(*[
                    self.cp(
                        getattr(subdir, "path", f"{source}/{getattr(subdir, 'path', '')}"),
                        f"{destination}/{getattr(subdir, 'path', '')}"
                    ) for subdir in batch
                ])
            # Process files in batches of 10
            files = getattr(data, "files", []) or []
            for i in range(0, len(files), 10):
                batch = files[i:i+10]
                await asyncio.gather(*[
                    self.write(
                        f"{destination}/{getattr(file, 'path', '')}",
                        await self.read(getattr(file, "path", f"{source}/{getattr(file, 'path', '')}"))
                    ) for file in batch
                ])
            return {
                "message": "Directory copied successfully",
                "source": source,
                "destination": destination,
            }
        raise Exception("Unsupported file type")

    def format_path(self, path: str) -> str:
        if path == "/":
            return "%2F"
        if path.startswith("/"):
            path = path[1:]
        return path


import argparse
import os
import sys
import asyncio
import requests


from typing import Any, Dict, Callable, Awaitable
from logging import getLogger
from ..client import client
from ..common.env import env
from ..common.internal import get_global_unique_hash
from ..common.settings import settings
from ..instrumentation.span import SpanManager



class BlJobWrapper:
    def get_arguments(self) -> Dict[str, Any]:
        if not os.getenv('BL_EXECUTION_DATA_URL'):
            parser = argparse.ArgumentParser()
            # Parse known args, ignore unknown
            args, unknown = parser.parse_known_args()
            # Convert to dict and include unknown args
            args_dict = vars(args)
            # Add unknown args to dict
            for i in range(0, len(unknown), 2):
                if i + 1 < len(unknown):
                    key = unknown[i].lstrip('-')
                    args_dict[key] = unknown[i + 1]
            return args_dict

        response = requests.get(os.getenv('BL_EXECUTION_DATA_URL'))
        data = response.json()
        tasks = data.get('tasks', [])
        return tasks[self.index] if self.index < len(tasks) else {}

    @property
    def index_key(self) -> str:
        return os.getenv('BL_TASK_KEY', 'TASK_INDEX')

    @property
    def index(self) -> int:
        index_value = os.getenv(self.index_key)
        return int(index_value) if index_value else 0

    def start(self, func: Callable):
        """
        Run a job defined in a function, it's run in the current process.
        Handles both async and sync functions.
        Arguments are passed as keyword arguments to the function.
        """
        attributes = {
            "span.type": "job.start",
        }
        try:
            parsed_args = self.get_arguments()
            if asyncio.iscoroutinefunction(func):
                asyncio.run(func(**parsed_args))
            else:
                func(**parsed_args)
        except Exception as error:
            print('Job execution failed:', error, file=sys.stderr)
            sys.exit(1)




logger = getLogger(__name__)

class BlJob:
    def __init__(self, name: str):
        self.name = name

    @property
    def internal_url(self):
        """Get the internal URL for the job using a hash of workspace and job name."""
        hash = get_global_unique_hash(settings.workspace, "job", self.name)
        return f"{settings.run_internal_protocol}://bl-{settings.env}-{hash}.{settings.run_internal_hostname}"

    @property
    def forced_url(self):
        """Get the forced URL from environment variables if set."""
        env_var = self.name.replace("-", "_").upper()
        if env[f"BL_JOB_{env_var}_URL"]:
            return env[f"BL_JOB_{env_var}_URL"]
        return None

    @property
    def external_url(self):
        return f"{settings.run_url}/{settings.workspace}/jobs/{self.name}"

    @property
    def fallback_url(self):
        if self.external_url != self.url:
            return self.external_url
        return None

    @property
    def url(self):
        if self.forced_url:
            return self.forced_url
        if settings.run_internal_hostname:
            return self.internal_url
        return self.external_url

    def call(self, url, input_data, headers: dict = {}, params: dict = {}):
        body = {
            "tasks": input_data
        }

        return client.get_httpx_client().post(
            url+"/executions",
            headers={
                'Content-Type': 'application/json',
                **headers
            },
            json=body,
            params=params
        )

    async def acall(self, url, input_data, headers: dict = {}, params: dict = {}):
        logger.debug(f"Job Calling: {self.name}")
        body = {
            "tasks": input_data
        }

        return await client.get_async_httpx_client().post(
            url+"/executions",
            headers={
                'Content-Type': 'application/json',
                **headers
            },
            json=body,
            params=params
        )

    def run(self, input: Any, headers: dict = {}, params: dict = {}) -> str:
        attributes = {
            "job.name": self.name,
            "span.type": "job.run",
        }
        with SpanManager("blaxel-tracer").create_active_span(self.name, attributes) as span:
            logger.debug(f"Job Calling: {self.name}")
            response = self.call(self.url, input, headers, params)
            if response.status_code >= 400:
                if not self.fallback_url:
                    span.set_attribute("job.run.error", response.text)
                    raise Exception(f"Job {self.name} returned status code {response.status_code} with body {response.text}")
                response = self.call(self.fallback_url, input, headers, params)
                if response.status_code >= 400:
                    span.set_attribute("job.run.error", response.text)
                    raise Exception(f"Job {self.name} returned status code {response.status_code} with body {response.text}")
            span.set_attribute("job.run.result", response.text)
            return response.text

    async def arun(self, input: Any, headers: dict = {}, params: dict = {}) -> Awaitable[str]:
        attributes = {
            "job.name": self.name,
            "span.type": "job.run",
        }
        with SpanManager("blaxel-tracer").create_active_span(self.name, attributes) as span:
            logger.debug(f"Job Calling: {self.name}")
            response = await self.acall(self.url, input, headers, params)
            if response.status_code >= 400:
                if not self.fallback_url:
                    span.set_attribute("job.run.error", response.text)
                    raise Exception(f"Job {self.name} returned status code {response.status_code} with body {response.text}")
                response = await self.acall(self.fallback_url, input, headers, params)
                if response.status_code >= 400:
                    span.set_attribute("job.run.error", response.text)
                    raise Exception(f"Job {self.name} returned status code {response.status_code} with body {response.text}")
            span.set_attribute("job.run.result", response.text)
            return response.text

    def __str__(self):
        return f"Job {self.name}"

    def __repr__(self):
        return self.__str__()



def bl_job(name: str):
    return BlJob(name)

# Create a singleton instance
bl_start_job = BlJobWrapper()



import asyncio

from dotenv import load_dotenv

load_dotenv()

from logging import getLogger

from pydantic_ai import Agent, CallToolsNode
from pydantic_ai.messages import ToolCallPart
from pydantic_ai.models import ModelSettings

from blaxel.models import bl_model
from blaxel.tools import bl_tools

logger = getLogger(__name__)

# MODEL = "gpt-4o-mini"
# MODEL = "claude-3-5-sonnet"
# MODEL = "xai-grok-beta"
# MODEL = "cohere-command-r-plus" # x -> tool call not supported
MODEL = "gemini-2-0-flash"
# MODEL = "deepseek-chat"
# MODEL = "mistral-large-latest"

async def main():
    tools = await bl_tools(["blaxel-search"]).to_pydantic()

    model = await bl_model(MODEL).to_pydantic()

    agent = Agent(model=model, tools=tools, model_settings=ModelSettings(temperature=0))
    async with agent.iter('Hello') as agent_run:
        async for node in agent_run:
            if isinstance(node, CallToolsNode):
                for part in node.model_response.parts:
                    if isinstance(part, ToolCallPart):
                        logger.info(f"Tool call: {part}")
                    else:
                        logger.info(f"Response: {part}")

if __name__ == "__main__":
    asyncio.run(main())



import nest_asyncio

nest_asyncio.apply()
import asyncio

from dotenv import load_dotenv

load_dotenv()

from logging import getLogger

from crewai import Agent, Crew, Task

from blaxel.models import bl_model
from blaxel.tools import bl_tools

logger = getLogger(__name__)

async def main():
    tools = await bl_tools(["blaxel-search"]).to_crewai()
    model = await bl_model("gpt-4o-mini").to_crewai()

    agent = Agent(
        role="Weather Researcher",
        goal="Find the weather in a specific city",
        backstory="You are an experienced weather researcher with attention to detail",
        llm=model,
        tools=tools,
        verbose=True,
    )

    my_crew = Crew(
        agents=[agent],
        tasks=[
            Task(
                description="Find the weather in a specific city",
                expected_output="Weather in San francisco.",
                agent=agent
            )
        ],
        verbose=True,
    )

    result = my_crew.kickoff()
    logger.info(result.raw)

if __name__ == "__main__":
    asyncio.run(main())



from livekit.agents import function_tool, llm

from .types import Tool


def livekit_coroutine(tool: Tool):
    async def livekit_coroutine_wrapper(raw_arguments: dict[str, object]):
        result = await tool.coroutine(**raw_arguments)
        return result.model_dump_json()
    return livekit_coroutine_wrapper

def get_livekit_tools(tools: list[Tool]) -> list[llm.FunctionTool]:
    livekit_tools = []
    for tool in tools:
        livekit_tools.append(function_tool(
            livekit_coroutine(tool),
            raw_schema={
                "name": tool.name,
                "description": tool.description,
                "parameters": tool.input_schema,
            },
        ))
    return livekit_tools



from logging import getLogger

from langchain_anthropic import ChatAnthropic
from langchain_cerebras import ChatCerebras
from langchain_cohere import ChatCohere
from langchain_deepseek import ChatDeepSeek
from langchain_openai import ChatOpenAI
from langchain_xai import ChatXAI

from ..common.settings import settings
from .custom.langchain.gemini import ChatGoogleGenerativeAI

logger = getLogger(__name__)

async def get_langchain_model(url: str,type: str, model: str, **kwargs):
    if type == 'mistral':
        return ChatOpenAI(
            api_key=settings.auth.token,
            model=model,
            base_url=f"{url}/v1",
            **kwargs
        )
    elif type == 'cohere':
        return ChatCohere(
            cohere_api_key=settings.auth.token,
            model=model,
            base_url=url,
            **kwargs
        )
    elif type == 'xai':
        return ChatXAI(
            model=model,
            api_key=settings.auth.token,
            xai_api_base=f"{url}/v1",
            **kwargs
        )
    elif type == 'deepseek':
        return ChatDeepSeek(
            api_key=settings.auth.token,
            model=model,
            api_base=f"{url}/v1",
            **kwargs
        )
    elif type == 'anthropic':
        return ChatAnthropic(
            api_key=settings.auth.token,
            anthropic_api_url=url,
            model=model,
            default_headers=settings.auth.get_headers(),
            **kwargs
        )
    elif type == 'gemini':
        return ChatGoogleGenerativeAI(
            model=model,
            client_options={"api_endpoint": url},
            additional_headers=settings.auth.get_headers(),
            transport="rest",
            **kwargs
        )
    elif type == "cerebras":
        return ChatCerebras(
            api_key=settings.auth.token,
            model=model,
            base_url=f"{url}/v1",
            **kwargs
        )
    else:
        if type != "openai":
            logger.warning(f"Model {model} is not supported by Langchain, defaulting to OpenAI")
        return ChatOpenAI(
            api_key=settings.auth.token,
            model=model,
            base_url=f"{url}/v1",
            **kwargs
        )



from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.integration_endpoint_token import IntegrationEndpointToken


T = TypeVar("T", bound="IntegrationEndpoint")


@_attrs_define
class IntegrationEndpoint:
    """Integration endpoint

    Attributes:
        body (Union[Unset, str]): Integration endpoint body
        ignore_models (Union[Unset, list[Any]]): Integration endpoint ignore models
        method (Union[Unset, str]): Integration endpoint method
        models (Union[Unset, list[Any]]): Integration endpoint models
        stream_key (Union[Unset, str]): Integration endpoint stream key
        stream_token (Union[Unset, IntegrationEndpointToken]): Integration endpoint token
        token (Union[Unset, IntegrationEndpointToken]): Integration endpoint token
    """

    body: Union[Unset, str] = UNSET
    ignore_models: Union[Unset, list[Any]] = UNSET
    method: Union[Unset, str] = UNSET
    models: Union[Unset, list[Any]] = UNSET
    stream_key: Union[Unset, str] = UNSET
    stream_token: Union[Unset, "IntegrationEndpointToken"] = UNSET
    token: Union[Unset, "IntegrationEndpointToken"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        body = self.body

        ignore_models: Union[Unset, list[Any]] = UNSET
        if not isinstance(self.ignore_models, Unset):
            ignore_models = self.ignore_models

        method = self.method

        models: Union[Unset, list[Any]] = UNSET
        if not isinstance(self.models, Unset):
            models = self.models

        stream_key = self.stream_key

        stream_token: Union[Unset, dict[str, Any]] = UNSET
        if self.stream_token and not isinstance(self.stream_token, Unset) and not isinstance(self.stream_token, dict):
            stream_token = self.stream_token.to_dict()
        elif self.stream_token and isinstance(self.stream_token, dict):
            stream_token = self.stream_token

        token: Union[Unset, dict[str, Any]] = UNSET
        if self.token and not isinstance(self.token, Unset) and not isinstance(self.token, dict):
            token = self.token.to_dict()
        elif self.token and isinstance(self.token, dict):
            token = self.token

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if body is not UNSET:
            field_dict["body"] = body
        if ignore_models is not UNSET:
            field_dict["ignoreModels"] = ignore_models
        if method is not UNSET:
            field_dict["method"] = method
        if models is not UNSET:
            field_dict["models"] = models
        if stream_key is not UNSET:
            field_dict["streamKey"] = stream_key
        if stream_token is not UNSET:
            field_dict["streamToken"] = stream_token
        if token is not UNSET:
            field_dict["token"] = token

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.integration_endpoint_token import IntegrationEndpointToken

        if not src_dict:
            return None
        d = src_dict.copy()
        body = d.pop("body", UNSET)

        ignore_models = cast(list[Any], d.pop("ignoreModels", UNSET))

        method = d.pop("method", UNSET)

        models = cast(list[Any], d.pop("models", UNSET))

        stream_key = d.pop("streamKey", UNSET)

        _stream_token = d.pop("streamToken", UNSET)
        stream_token: Union[Unset, IntegrationEndpointToken]
        if isinstance(_stream_token, Unset):
            stream_token = UNSET
        else:
            stream_token = IntegrationEndpointToken.from_dict(_stream_token)

        _token = d.pop("token", UNSET)
        token: Union[Unset, IntegrationEndpointToken]
        if isinstance(_token, Unset):
            token = UNSET
        else:
            token = IntegrationEndpointToken.from_dict(_token)

        integration_endpoint = cls(
            body=body,
            ignore_models=ignore_models,
            method=method,
            models=models,
            stream_key=stream_key,
            stream_token=stream_token,
            token=token,
        )

        integration_endpoint.additional_properties = d
        return integration_endpoint

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="PendingInvitationWorkspaceDetails")


@_attrs_define
class PendingInvitationWorkspaceDetails:
    """Workspace details

    Attributes:
        emails (Union[Unset, list[Any]]): List of user emails in the workspace
        user_number (Union[Unset, float]): Number of users in the workspace
    """

    emails: Union[Unset, list[Any]] = UNSET
    user_number: Union[Unset, float] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        emails: Union[Unset, list[Any]] = UNSET
        if not isinstance(self.emails, Unset):
            emails = self.emails

        user_number = self.user_number

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if emails is not UNSET:
            field_dict["emails"] = emails
        if user_number is not UNSET:
            field_dict["user_number"] = user_number

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        emails = cast(list[Any], d.pop("emails", UNSET))

        user_number = d.pop("user_number", UNSET)

        pending_invitation_workspace_details = cls(
            emails=emails,
            user_number=user_number,
        )

        pending_invitation_workspace_details.additional_properties = d
        return pending_invitation_workspace_details

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import traceback
from logging import getLogger
from time import time

from fastapi.responses import JSONResponse

logger = getLogger(__name__)
total = time()
start = time()

from contextlib import asynccontextmanager

import uvicorn
from asgi_correlation_id import CorrelationIdMiddleware
from fastapi import FastAPI, Request, Response, status
from fastapi.encoders import jsonable_encoder
from langchain_core.messages import HumanMessage
from langgraph.prebuilt import create_react_agent
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor

from blaxel.instrumentation.span import SpanManager
from blaxel.models import bl_model
from blaxel.tools import bl_tools

logger.info(f"Loaded blaxel in {round(time() - start, 4)} seconds")

@asynccontextmanager
async def lifespan(app: FastAPI):
    logger.info("Server running on port 1338")
    yield
    logger.info("Server shutting down")

start = time()
app = FastAPI(lifespan=lifespan)
logger.info(f"Created app in {round(time() - start, 4)} seconds")
start = time()
app.add_middleware(CorrelationIdMiddleware)
logger.info(f"Added correlation id middleware in {round(time() - start, 4)} seconds")

@app.middleware("http")
async def log_requests(request: Request, call_next):
    start_time = time()

    response: Response = await call_next(request)

    process_time = (time() - start_time) * 1000
    formatted_process_time = f'{process_time:.2f}'
    rid_header = response.headers.get("X-Request-Id")
    request_id = rid_header or response.headers.get("X-Blaxel-Request-Id")
    logger.info(f"{request.method} {request.url.path} {response.status_code} {formatted_process_time}ms rid={request_id}")

    return response

@app.exception_handler(Exception)
async def validation_exception_handler(request: Request, e: Exception):
    logger.error(f"Error on request {request.method} {request.url.path}: {e}")

    # Get the full traceback information
    tb_str = traceback.format_exception(type(e), e, e.__traceback__)
    logger.error(f"Stacktrace: {''.join(tb_str)}")
    return JSONResponse(
        status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
        content=jsonable_encoder({"detail": str(e)}),
    )


@app.post("/")
async def handle_request(request: Request):
    start = time()
    tools = await bl_tools(["blaxel-search"]).to_langchain()
    logger.info(f"Loaded tools in {round(time() - start, 4)} seconds")
    start = time()
    logger.info(f"Converted tools to langchain in {round(time() - start, 4)} seconds")
    start = time()
    model = await bl_model("gpt-4o-mini").to_langchain()
    logger.info(f"Retrieved model in langchain in {round(time() - start, 4)} seconds")
    start = time()
    agent = create_react_agent(model=model, tools=tools, prompt="You are a helpful assistant that can answer questions and help with tasks.")
    logger.info(f"Create agent in {round(time() - start, 4)} seconds")
    with SpanManager("blaxel-langchain").create_active_span("agent-request", {}) as span:
        body = await request.json()
        result = await agent.ainvoke({"messages": [HumanMessage(body.get("inputs", ""))]})
        span.set_attribute("agent.result", result["messages"][-1].content)

        return Response(
            result["messages"][-1].content,
            media_type="text/plain"
        )

start = time()
FastAPIInstrumentor.instrument_app(app)
logger.info(f"Instrumented app in {round(time() - start, 4)} seconds")

if __name__ == "__main__":
    uvicorn.run(app, host="0.0.0.0", port=1338, log_level="critical", loop="asyncio")


import os

from httpx import Response

from ..client.models import Sandbox
from ..common.internal import get_global_unique_hash
from ..common.settings import settings
from .client.client import client
from .client.models import ErrorResponse


class ResponseError(Exception):
    def __init__(self, response: Response):
        self.status_code = response.status_code
        self.status_text = response.content
        self.error = None
        data_error = {
            "status": response.status_code,
            "statusText": response.content,
        }
        if hasattr(response, "parsed") and isinstance(response.parsed, ErrorResponse):
            data_error["error"] = response.parsed.error
            self.error = response.parsed.error
        super().__init__(str(data_error))


class SandboxHandleBase:
    def __init__(self, sandbox: Sandbox):
        self.sandbox = sandbox
        self.client = client.with_base_url(self.url).with_headers(settings.headers)

    @property
    def name(self):
        return self.sandbox.metadata and self.sandbox.metadata.name

    @property
    def fallback_url(self):
        if self.external_url != self.url:
            return self.external_url
        return None

    @property
    def external_url(self):
        return f"{settings.run_url}/{settings.workspace}/sandboxes/{self.name}"

    @property
    def internal_url(self):
        hash_ = get_global_unique_hash(settings.workspace, "sandbox", self.name)
        return f"{settings.run_internal_protocol}://bl-{settings.env}-{hash_}.{settings.run_internal_hostname}"

    @property
    def forced_url(self):
        env_var = self.name.replace("-", "_").upper()
        env_name = f"BL_SANDBOX_{env_var}_URL"
        return os.environ.get(env_name)

    @property
    def url(self):
        if self.forced_url:
            return self.forced_url
        # Uncomment and use this when agent and mcp are available in mk3
        # Update all requests made in this package to use fallbackUrl when internalUrl is not working
        # if settings.run_internal_hostname:
        #     return self.internal_url
        return self.external_url

    def handle_response(self, response: Response):
        if response.status_code >= 400:
            raise ResponseError(response)




from pydantic_ai import RunContext
from pydantic_ai.tools import Tool as PydanticTool
from pydantic_ai.tools import ToolDefinition

from .types import Tool


def get_pydantic_tool(tool: Tool) -> PydanticTool:
    """
    Converts a custom Tool object into a Pydantic AI Tool object.

    Uses the attributes defined in the custom Tool (name, description, input_schema)
    by defining a dynamic 'prepare' function that modifies the ToolDefinition
    generated by Pydantic AI.
    """
    # Select the function to use, prioritizing the async version
    func = tool.coroutine if tool.coroutine else tool.sync_coroutine
    if not func:
        raise ValueError(
            f"Tool '{tool.name}' must have either a coroutine or sync_coroutine defined."
        )

    # Define the prepare function dynamically to capture the 'tool' object
    # Assuming RunContext and ToolDefinition types based on the pydantic_ai example
    async def prepare_tool(
        ctx: RunContext, tool_def: ToolDefinition
    ) -> ToolDefinition | None:
        """Dynamically prepares the ToolDefinition using the custom Tool's attributes."""
        tool_def.name = tool.name  # Override inferred name
        tool_def.description = tool.description  # Override inferred description
        tool_def.parameters_json_schema = tool.input_schema
        return tool_def

    async def pydantic_function(**kwargs):
        return await func(**kwargs)

    # Create the Pydantic AI Tool, passing the function and the prepare hook
    return PydanticTool(
        pydantic_function,
        name=tool.name,
        description=tool.description,
        prepare=prepare_tool,
        takes_ctx=False,
    )


def get_pydantic_tools(tools: list[Tool]) -> list[PydanticTool]:
    return [get_pydantic_tool(tool) for tool in tools]


import asyncio

from dotenv import load_dotenv

load_dotenv()

from logging import getLogger

from langchain_core.messages import AIMessage
from langgraph.prebuilt import create_react_agent

from blaxel.models import bl_model
from blaxel.tools import bl_tools

logger = getLogger(__name__)

MODEL = "gpt-4o-mini"
# MODEL = "claude-3-5-sonnet"
# MODEL = "xai-grok-beta"
# MODEL = "cohere-command-r-plus" # x -> tool call not supported
# MODEL = "gemini-2-0-flash"
# MODEL = "deepseek-chat"
# MODEL = "mistral-large-latest"

async def main():
    tools = await bl_tools(["blaxel-search"]).to_langchain()

    model = await bl_model(MODEL, temperature=0).to_langchain()

    agent = create_react_agent(model=model, tools=tools, prompt="You are a helpful assistant. Maximum number of tool call is 1")
    input = "Search online for the current weather in San Francisco ?"
    # input = "What are the tools in your arsenal ?"
    # input = "Hello world"
    messages = {"messages": [("user", input)]}
    responses = []
    async for message in agent.astream(messages):
        if "agent" in message and len(message["agent"].get("messages", [])) > 0:
            msg = message["agent"].get("messages", [])[0]
            if isinstance(msg, AIMessage):
                if msg.additional_kwargs:
                    if msg.additional_kwargs.get("tool_calls"):
                        for tool_call in msg.additional_kwargs.get("tool_calls"):
                            fn = tool_call.get("function", {})
                            if fn:
                                logger.info(f"Tool request: {fn['name']} with arguments {fn['arguments']}")
        responses.append(message)

    logger.info(responses[-1]["agent"]["messages"][-1].content)

if __name__ == "__main__":
    asyncio.run(main())